@@ -1,5 +1,5 @@
 /* dialog.c --- the password dialog and splash screen.
- * xscreensaver, Copyright © 1993-2021 Jamie Zawinski <jwz@jwz.org>
+ * xscreensaver, Copyright © 1993-2025 Jamie Zawinski <jwz@jwz.org>
  *
  * Permission to use, copy, modify, distribute, and sell this software and its
  * documentation for any purpose is hereby granted without fee, provided that
@@ -47,6 +47,10 @@
 #include <ctype.h>
 #include <pwd.h>
 
+#ifndef HAVE_XINPUT
+# error The XInput2 extension is required
+#endif
+
 #include <X11/Xproto.h>		/* for CARD32 */
 #include <X11/Xlib.h>
 #include <X11/Xatom.h>
@@ -129,7 +133,7 @@ typedef struct {
 typedef struct {
   char *text;
   XftFont *font;
-  XftColor fg;
+  XftColor fg, fg2;
   Pixel bg;
   enum { LABEL, BUTTON, TEXT, TEXT_RO } type;
   line_align align;
@@ -150,10 +154,12 @@ struct window_state {
   Window window;
   Colormap cmap;
 
-  Bool splash_p;
+  int splash_p;
   auth_state auth_state;
   int xi_opcode;
   int xkb_opcode;
+  XIM im;
+  XIC ic;
 
   /* Variant strings
    */
@@ -181,6 +187,7 @@ struct window_state {
   XtIntervalId timer;
 
   XtIntervalId cursor_timer;		/* Blink the I-beam */
+  XtIntervalId bs_timer;		/* Auto-repeat Backspace */
   int i_beam;
 
   double start_time, end_time;
@@ -214,6 +221,7 @@ struct window_state {
   XftColor xft_foreground;
   XftColor xft_text_foreground;
   XftColor xft_button_foreground;
+  XftColor xft_button_disabled;
   XftColor xft_error_foreground;
   Pixel passwd_background;
   Pixel thermo_foreground;
@@ -489,10 +497,13 @@ draw_dialog_line (window_state *ws, Drawable d, dialog_line *line,
 
   if (line->type == BUTTON &&
       line->button &&
-      (line->button->down_p || line->button->disabled_p))
+      line->button->down_p)
     xoff2 = yoff2 = MIN (ws->shadow_width, line->font->ascent/2);
 
-  XftDrawStringUtf8_multi (ws->xftdraw, &line->fg, line->font,
+  XftDrawStringUtf8_multi (ws->xftdraw, 
+                           (line->button && line->button->disabled_p
+                            ? &line->fg2 : &line->fg),
+                           line->font,
                            left + xoff2,
                            y + ypad + yoff2 + line->font->ascent,
                            (FcChar8 *) text, strlen (text),
@@ -672,6 +683,31 @@ get_xft_color (window_state *ws, XftColor *ret,
                      s, ret);
 }
 
+static void
+dim_xft_color (window_state *ws, const XftColor *in, Pixel bg, XftColor *out)
+{
+  double dim = 0.6;
+# if 0  /* Turns out Xft alpha doesn't work. How very. */
+  XRenderColor rc = in->color;
+  rc.alpha *= dim;
+# else
+  XRenderColor rc;
+  XColor xc;
+  xc.pixel = bg;
+  XQueryColor (ws->dpy, DefaultColormapOfScreen (ws->screen), &xc);
+  rc.red   = dim * in->color.red   + (1-dim) * xc.red;
+  rc.green = dim * in->color.green + (1-dim) * xc.green;
+  rc.blue  = dim * in->color.blue  + (1-dim) * xc.blue;
+  rc.alpha = in->color.alpha;
+# endif
+  if (! XftColorAllocValue (ws->dpy,
+                            DefaultVisualOfScreen(ws->screen),
+                            DefaultColormapOfScreen (ws->screen),
+                            &rc, out))
+    abort();
+}
+
+
 static int
 get_int (window_state *ws, const char *name, const char *rclass)
 {
@@ -692,10 +728,12 @@ choose_asterisk (window_state *ws)
   int i, L;
   for (i = 0; i < countof (candidates) - 1; i++)
     {
+# ifdef HAVE_XFT
       if (XftCharExists (ws->dpy, ws->label_font, (FcChar32) *uc))
         break;
       if (debug_p)
         fprintf (stderr, "%s: char U+%0lX does not exist\n", blurb(), *uc);
+# endif
       uc++;
     }
 
@@ -844,6 +882,40 @@ grab_keyboard_and_mouse (window_state *ws)
 }
 
 
+/* In the olden days, the way to have multiple keyboard layouts was:
+
+   A) Run "xmodmap" to load a different layout, one at a time.  Fun!
+
+   B) Have a "Mode_switch" key somewhere on your keyboard (traditionally on
+      the key labeled "AltGr"), and attached to Mod5.  When "Mode_switch" was
+      toggled on, the KeyPress event's 'state' bitmask would have the Mod5 bit
+      set, meaning that the client should use the keycode's 3rd and 4th
+      columns in the map instead of the 1st and 2nd.  In this way you could
+      have exactly two layouts.
+
+   But in This Modern World, the XKeyboard extension does... other magic.
+
+   C) Under LXDE circa 2021-2025, the Keyboard Preferences dialog lets you
+      select one alternate keyboard, and a key combo, typically Ctrl + Alt.
+      Through some mechanism I do not understand, it configures XKB so that
+      pressing Ctrl + Alt causes XLookupString to map the keycode to a
+      different keysym.  It does this without the 'state' field having Mod5
+      set.  
+
+      Bit 13 is set in 'state', but I don't know what that is or where it is
+      defined.  If I had to guess, I'd say that XKB.h:XkbBuildCoreState()
+      uses bits 13 and 14 to indicate up to 4 alternate keyboard layouts, but
+      I have't found any documentation to this effect, nor do I know how
+      XLookupString knows to interpret those bits.  Or why LXDE limits you to
+      just 1 alternate layout.
+
+      When you press Ctrl + Alt, an XkbEvent comes in, which is our clue that
+      the keyboard layout *may* have changed.  When that happens, we call
+      this function to present the name of the current layout on the window.
+
+      Pressing Ctrl + Alt does not send a MappingNotify event, because that
+      would make too much sense.
+ */
 static void
 get_keyboard_layout (window_state *ws)
 {
@@ -944,6 +1016,24 @@ create_window (window_state *ws, int w, int h)
   XSetWindowColormap (ws->dpy, ws->window, ws->cmap);
   xscreensaver_set_wm_atoms (ws->dpy, ws->window, w, h, 0);
 
+  /* An input method is necessary for dead keys to work.
+   */
+  if (! ws->im)
+    ws->im = XOpenIM (ws->dpy, NULL, NULL, NULL);
+  if (ws->ic)
+    XDestroyIC (ws->ic), ws->ic = 0;
+  if (ws->im)
+    ws->ic = XCreateIC (ws->im,
+                        XNInputStyle, XIMPreeditNothing | XIMStatusNothing,
+                        XNClientWindow, ws->window,
+                        NULL);
+  if (ws->ic)
+    XSetICFocus (ws->ic);
+
+  if (verbose_p)
+    fprintf (stderr, "%s: %s input method\n", blurb(),
+             ws->ic ? "attached" : "failed to attach");
+
   if (ow)
     {
       XMapRaised (ws->dpy, ws->window);
@@ -955,7 +1045,7 @@ create_window (window_state *ws, int w, int h)
 /* Loads resources and creates and returns the global window state.
  */
 static window_state *
-window_init (Widget root_widget, Bool splash_p)
+window_init (Widget root_widget, int splash_p)
 {
   Display *dpy = XtDisplay (root_widget);
   window_state *ws;
@@ -1042,6 +1132,8 @@ window_init (Widget root_widget, Bool splash_p)
                  "error.foreground", "Error.Foreground");
   get_xft_color (ws, &ws->xft_button_foreground,
                  "button.foreground", "Button.Foreground");
+  dim_xft_color (ws, &ws->xft_button_foreground, ws->background,
+                 &ws->xft_button_disabled);
 
   ws->shadow_top    = get_color (ws, "topShadowColor", "Foreground");
   ws->shadow_bottom = get_color (ws, "bottomShadowColor", "Background");
@@ -1328,27 +1420,111 @@ window_draw (window_state *ws)
   lines[i].text  = ws->heading_label;			/* XScreenSaver */
   lines[i].font  = ws->heading_font;
   lines[i].fg    = ws->xft_foreground;
+  lines[i].fg2   = lines[i].fg;
   lines[i].bg    = ws->background;
   lines[i].type  = LABEL;
   lines[i].align = CENTER;
   i++;
 
+  /* If you are in here because you're planning on disabling this notice
+     before redistributing my software, please don't.
+
+     I sincerely request that you do one of the following:
+
+         1: leave this code intact and this warning in place, -OR-
+
+         2: Remove xscreensaver from your distribution.
+
+     I would seriously prefer that you not distribute my software at all
+     than that you distribute one version and then never update it for
+     years.
+
+     I am *constantly* getting email from users reporting bugs that have
+     been fixed for literally years who have no idea that the software
+     they are running is years out of date.  Yes, it would be great if we
+     lived in the ideal world where people checked that they were running
+     the latest release before they report a bug, but we don't.  To most
+     people, "running the latest release" is synonymous with "running the
+     latest release that my distro packages for me."
+
+     When they even bother to tell me what version they're running, I
+     say, "That version is three years old!", and they say "But this is
+     the latest version my distro ships".  Then I say, "your distro
+     sucks", and they say "but I don't know how to compile from source,
+     herp derp I eat paste", and *everybody* goes away unhappy.
+
+     It wastes an enormous amount of my time, but worse than that, it
+     does a grave disservice to the users, who are stuck experiencing
+     bugs that are already fixed!  These users think they are running the
+     latest release, and they are not.  They would like to be running the
+     actual latest release, but they don't know how, because their distro
+     makes that very difficult for them.  It's terrible for everyone, and
+     kind of makes me regret ever having released this software in the
+     first place.
+
+     So seriously. I ask that if you're planning on disabling this
+     obsolescence warning, that you instead just remove xscreensaver from
+     your distro entirely.  Everybody will be happier that way.  Check
+     out gnome-screensaver instead, I understand it's really nice.
+
+     Of course, my license allows you to ignore me and do whatever the
+     fuck you want, but as the author, I hope you will have the common
+     courtesy of complying with my request.
+
+     Thank you!
+
+     jwz, 2014, 2016, 2018, 2021.
+
+     PS: In particular, since Debian refuses to upgrade software on any
+     kind of rational timeline, I have asked that they stop shipping
+     xscreensaver at all.  They have refused.  Instead of upgrading the
+     software, they simply patched out this warning.
+
+     If you want to witness the sad state of the open source peanut
+     gallery, look no farther than the comments on my blog:
+     http://jwz.org/b/yiYo
+
+     Many of these people fall back on their go-to argument of, "If it is
+     legal, it must be right."  If you believe in that rhetorical device
+     then you are a terrible person, and possibly a sociopath.
+
+     There are also the armchair lawyers who say "Well, instead of
+     *asking* people to do the right thing out of common courtesy, you
+     should just change your license to prohibit them from acting
+     amorally."  Again, this is the answer of a sociopath, but that aside,
+     if you devote even a second's thought to this you will realize that
+     the end result of this would be for distros like Debian to just keep
+     shipping the last version with the old license and then never
+     upgrading it again -- which would be the worst possible outcome for
+     everyone involved, most especially the users.
+
+     Also, some have incorrectly characterized this as a "time bomb".
+     It is a software update notification, nothing more.  A "time bomb"
+     makes software stop working.  This merely alerts the user that the
+     security-critical software that they are running is dangerously out
+     of date.
+
+     If you have read all of the above, and still decide to intentionally
+     disrespect the wishes of the person who wrote all of this software for
+     you -- you are a terrible person.  Kindly go fuck yourself.
+  */
   if (time ((time_t *) 0) - XSCREENSAVER_RELEASED > 60*60*24*30*17)
     {
       lines[i].text  = _("Update available!\nThis version is very old.\n");
       lines[i].font  = ws->error_font;
       lines[i].fg    = ws->xft_error_foreground;
+      lines[i].fg2   = lines[i].fg;
       lines[i].bg    = ws->background;
       lines[i].type  = LABEL;
       lines[i].align = CENTER;
       i++;
     }
-  else if (strstr (ws->version, "a") ||
-      strstr (ws->version, "b"))
+  else if (strstr (ws->version, "a") || strstr (ws->version, "b"))
     {
       lines[i].text  = _("PRE-RELEASE VERSION");
       lines[i].font  = ws->error_font;
       lines[i].fg    = ws->xft_error_foreground;
+      lines[i].fg2   = lines[i].fg;
       lines[i].bg    = ws->background;
       lines[i].type  = LABEL;
       lines[i].align = CENTER;
@@ -1360,6 +1536,7 @@ window_draw (window_state *ws)
       lines[i].text  = ws->hostname_label;
       lines[i].font  = ws->hostname_font;
       lines[i].fg    = ws->xft_foreground;
+      lines[i].fg2   = lines[i].fg;
       lines[i].bg    = ws->background;
       lines[i].type  = LABEL;
       lines[i].align = CENTER;
@@ -1370,6 +1547,7 @@ window_draw (window_state *ws)
   lines[i].text  = "";                 \
   lines[i].font  = ws->body_font;      \
   lines[i].fg    = ws->xft_foreground; \
+  lines[i].fg2   = lines[i].fg;        \
   lines[i].bg    = ws->background;     \
   lines[i].type  = LABEL;              \
   lines[i].align = CENTER;             \
@@ -1383,6 +1561,7 @@ window_draw (window_state *ws)
         _("DEBUG MODE:\nAll keystrokes are being logged to stderr.\n");
       lines[i].font  = ws->error_font;
       lines[i].fg    = ws->xft_error_foreground;
+      lines[i].fg2   = lines[i].fg;
       lines[i].bg    = ws->background;
       lines[i].type  = LABEL;
       lines[i].align = CENTER;
@@ -1394,6 +1573,7 @@ window_draw (window_state *ws)
       lines[i].text  = ws->body_label;	    /* Copyright or error message */
       lines[i].font  = ws->body_font;
       lines[i].fg    = ws->xft_foreground;
+      lines[i].fg2   = lines[i].fg;
       lines[i].bg    = ws->background;
       lines[i].type  = LABEL;
       lines[i].align = CENTER;
@@ -1414,6 +1594,7 @@ window_draw (window_state *ws)
         lines[i].fg    = (ws->msgs[j].type == AUTH_MSGTYPE_ERROR
                           ? ws->xft_error_foreground
                           : ws->xft_foreground);
+        lines[i].fg2   = lines[i].fg;
         lines[i].bg    = ws->background;
         lines[i].type  = LABEL;
         lines[i].align = CENTER;
@@ -1429,6 +1610,7 @@ window_draw (window_state *ws)
             lines[i].text    = _("Username:");
             lines[i].font    = ws->label_font;
             lines[i].fg      = ws->xft_foreground;
+            lines[i].fg2     = lines[i].fg;
             lines[i].bg      = ws->background;
             lines[i].type    = LABEL;
             lines[i].align   = LEFT;
@@ -1438,6 +1620,7 @@ window_draw (window_state *ws)
             lines[i].text    = ws->user;		/* $USER */
             lines[i].font    = ws->label_font;
             lines[i].fg      = ws->xft_text_foreground;
+            lines[i].fg2     = lines[i].fg;
             lines[i].bg      = ws->passwd_background;
             lines[i].type    = TEXT_RO;
             lines[i].align   = RIGHT;
@@ -1447,6 +1630,7 @@ window_draw (window_state *ws)
         lines[i].text    = trim (ws->msgs[j].msg);	/* PAM prompt text */
         lines[i].font    = ws->label_font;
         lines[i].fg      = ws->xft_foreground;
+        lines[i].fg2     = lines[i].fg;
         lines[i].bg      = ws->background;
         lines[i].type    = LABEL;
         lines[i].align   = LEFT;
@@ -1462,6 +1646,7 @@ window_draw (window_state *ws)
                               : "");
         lines[i].font    = ws->label_font;
         lines[i].fg      = ws->xft_text_foreground;
+        lines[i].fg2     = lines[i].fg;
         lines[i].bg      = ws->passwd_background;
         lines[i].type    = TEXT;
         lines[i].align   = RIGHT;
@@ -1474,6 +1659,7 @@ window_draw (window_state *ws)
             lines[i].text   = date_text;
             lines[i].font   = ws->date_font;
             lines[i].fg     = ws->xft_foreground;
+            lines[i].fg2    = lines[i].fg;
             lines[i].bg     = ws->background;
             lines[i].type   = LABEL;
             lines[i].align  = RIGHT;
@@ -1486,6 +1672,7 @@ window_draw (window_state *ws)
             lines[i].text   = ws->kbd_layout_label;
             lines[i].font   = ws->date_font;
             lines[i].fg     = ws->xft_foreground;
+            lines[i].fg2    = lines[i].fg;
             lines[i].bg     = ws->background;
             lines[i].type   = LABEL;
             lines[i].align  = RIGHT;
@@ -1534,6 +1721,7 @@ window_draw (window_state *ws)
       lines[i].text  = _("Settings");
       lines[i].font  = ws->button_font;
       lines[i].fg    = ws->xft_button_foreground;
+      lines[i].fg2   = ws->xft_button_disabled;
       lines[i].bg    = ws->button_background;
       lines[i].type  = BUTTON;
       lines[i].align = LEFT;
@@ -1541,9 +1729,14 @@ window_draw (window_state *ws)
       lines[i].button = &ws->demo_button_state;
       i++;
 
+      if (ws->splash_p > 1)
+        /* Settings button is disabled with --splash --splash */
+        ws->demo_button_state.disabled_p = True;
+
       lines[i].text  = _("Help");
       lines[i].font  = ws->button_font;
       lines[i].fg    = ws->xft_button_foreground;
+      lines[i].fg2   = ws->xft_button_disabled;
       lines[i].bg    = ws->button_background;
       lines[i].type  = BUTTON;
       lines[i].align = RIGHT;
@@ -1557,6 +1750,7 @@ window_draw (window_state *ws)
           lines[i].text  = _("New Login");
           lines[i].font  = ws->button_font;
           lines[i].fg    = ws->xft_button_foreground;
+          lines[i].fg2   = ws->xft_button_disabled;
           lines[i].bg    = ws->button_background;
           lines[i].type  = BUTTON;
           lines[i].align = LEFT;
@@ -1568,6 +1762,7 @@ window_draw (window_state *ws)
       lines[i].text  = _("OK");
       lines[i].font  = ws->button_font;
       lines[i].fg    = ws->xft_button_foreground;
+      lines[i].fg2   = ws->xft_button_disabled;
       lines[i].bg    = ws->button_background;
       lines[i].type  = BUTTON;
       lines[i].align = RIGHT;
@@ -1772,6 +1967,11 @@ destroy_window (window_state *ws)
       XtRemoveTimeOut (ws->cursor_timer);
       ws->cursor_timer = 0;
     }
+  if (ws->bs_timer)
+    {
+      XtRemoveTimeOut (ws->bs_timer);
+      ws->bs_timer = 0;
+    }
 
   while (XCheckMaskEvent (ws->dpy, PointerMotionMask, &event))
     if (verbose_p)
@@ -1881,18 +2081,30 @@ persistent_auth_status_failure (window_state *ws,
 
   if (increment_p && clear_p) abort();
 
-  /* Read the old property so that we can increment it. */
+  /* Read the old property so that we can increment it.
+     Structure:
+
+     XScreenSaver 6.00, 2021:	XScreenSaver >= 6.11, 2025:
+
+	0: count		0: count
+	1: 32 bit time_t	1: 64 bit time_t hi
+				2: 64 bit time_t lo
+   */
   if (XGetWindowProperty (dpy, w, prop,
                           0, 999, False, XA_INTEGER,
                           &type, &format, &nitems, &bytesafter,
                           &dataP)
       == Success
       && type == XA_INTEGER
-      && nitems >= 2
+      && nitems >= 3
       && dataP)
     {
-      count = ((PROP32 *) dataP) [0];
-      tt    = ((PROP32 *) dataP) [1];  /* Y2038 bug: unsigned 32 bit time_t */
+      PROP32 *data = (PROP32 *) dataP;
+      count = data[0];
+      tt = (time_t)				/* 64 bit time_t */
+        ((((unsigned long) data[1] & 0xFFFFFFFFL) << 32) |
+          ((unsigned long) data[2] & 0xFFFFFFFFL));
+
       if (verbose_p)
         fprintf (stderr, "%s: previous auth failures: %d @ %lu\n",
                  blurb(), count, (unsigned long) tt);
@@ -1909,7 +2121,7 @@ persistent_auth_status_failure (window_state *ws,
     }
   else if (increment_p)
     {
-      PROP32 vv[2];
+      PROP32 vv[3];
       count++;
 
       /* Remember the time of the *oldest* failed login.  A failed login
@@ -1919,9 +2131,11 @@ persistent_auth_status_failure (window_state *ws,
       if (tt <= 0) tt = time ((time_t *) 0);
 
       vv[0] = (PROP32) count;
-      vv[1] = (PROP32) tt;
+      vv[1] = (PROP32) (((unsigned long) tt) >> 32);
+      vv[2] = (PROP32) (((unsigned long) tt) & 0xFFFFFFFFL);
+
       XChangeProperty (dpy, w, prop, XA_INTEGER, 32,
-                       PropModeReplace, (unsigned char *) vv, 2);
+                       PropModeReplace, (unsigned char *) vv, 3);
       if (verbose_p)
         fprintf (stderr, "%s: saved auth failure: %d @ %lu\n",
                  blurb(), count, (unsigned long) tt);
@@ -1932,9 +2146,10 @@ persistent_auth_status_failure (window_state *ws,
 }
 
 
+static void bs_timer (XtPointer, XtIntervalId *);
 
 static void
-handle_keypress (window_state *ws, XKeyEvent *event)
+handle_keypress (window_state *ws, XKeyEvent *event, Bool filter_p)
 {
   unsigned char decoded [MAX_BYTES_PER_CHAR * 10]; /* leave some slack */
   KeySym keysym = 0;
@@ -1956,6 +2171,48 @@ handle_keypress (window_state *ws, XKeyEvent *event)
   int decoded_size = XLookupString (event, (char *)decoded, sizeof(decoded),
                                     &keysym, &ws->compose_status);
 
+  /* But if we have an input method, we can unambiguously get UTF8.
+   */
+  if (ws->ic && event->type == KeyPress)
+    {
+      char decoded2[sizeof(decoded)];
+      Status s = 0;
+      KeySym keysym2 = 0;
+      int size2 = Xutf8LookupString (ws->ic, (XKeyPressedEvent *) event,
+                                     decoded2, sizeof(decoded2)-1,
+                                     &keysym2, &s);
+      decoded2[size2] = 0;
+
+      switch (s) {
+      case XLookupChars:		/* Set 'decoded2' to a UTF8 string */
+        if (*decoded2)
+          {
+            strcpy ((char *) decoded, (char *) decoded2);
+            decoded_size = size2;
+          }
+        break;
+      case XLookupKeySym:		/* Set 'keysym2' but not 'decoded2' */
+        if (keysym2)
+          keysym = keysym2;
+        break;
+      case XLookupBoth:			/* Set 'keysym2' and 'decoded2' */
+        if (keysym2)
+          keysym = keysym2;
+        if (*decoded2)
+          {
+            strcpy ((char *) decoded, (char *) decoded2);
+            decoded_size = size2;
+          }
+        break;
+      case XLookupNone:			/* No input yet */
+      case XBufferOverflow:		/* 'decoded2' was too small */
+        break;
+      default:
+        abort();
+        break;
+      }
+    }
+
   if (decoded_size > MAX_BYTES_PER_CHAR)
     {
       /* The multi-byte character returned is too large. */
@@ -1997,6 +2254,15 @@ handle_keypress (window_state *ws, XKeyEvent *event)
                   }
                 ws->plaintext_passwd_char_size[nchars-1] = 0;
               }
+
+            /* The XInput2 extension does not send auto-repeat KeyPress
+               events, and it annoys people that you can't hold down the
+               Backspace key to clear the line.  So clear the whole line
+               if the key is held down for a little while.  */
+            if (ws->bs_timer)
+              XtRemoveTimeOut (ws->bs_timer);
+            ws->bs_timer =
+              XtAppAddTimeOut (ws->app, 1000 * 0.6, bs_timer, (XtPointer) ws);
           }
           break;
 
@@ -2008,7 +2274,7 @@ handle_keypress (window_state *ws, XKeyEvent *event)
           ws->auth_state = AUTH_CANCEL;
           break;
 
-        case '\025': case '\030':			/* Erase line */
+        case '\025': case '\030':			/* Erase line ^U ^X */
           memset (ws->plaintext_passwd, 0, sizeof (ws->plaintext_passwd));
           memset (ws->plaintext_passwd_char_size, 0, 
                   sizeof (ws->plaintext_passwd_char_size));
@@ -2028,7 +2294,10 @@ handle_keypress (window_state *ws, XKeyEvent *event)
     SELF_INSERT:
       nbytes = strlen (ws->plaintext_passwd);
       nchars = strlen (ws->plaintext_passwd_char_size);
-      if (nchars + 1 >= sizeof (ws->plaintext_passwd_char_size)-1 ||
+
+      if (filter_p)
+        ;  /* Ignore: it is probably the start of a dead-key sequence. */
+      else if (nchars + 1 >= sizeof (ws->plaintext_passwd_char_size)-1 ||
           nbytes + decoded_size >= sizeof (ws->plaintext_passwd)-1)
         XBell (ws->dpy, 0);  /* overflow */
       else if (decoded_size == 0)
@@ -2092,7 +2361,7 @@ handle_button (window_state *ws, XEvent *xev, line_button_state *bs)
 
 
 static Bool
-handle_event (window_state *ws, XEvent *xev)
+handle_event (window_state *ws, XEvent *xev, Bool filter_p)
 {
   Bool refresh_p = False;
   switch (xev->xany.type) {
@@ -2101,7 +2370,7 @@ handle_event (window_state *ws, XEvent *xev)
       ws->auth_state = AUTH_CANCEL;
     else
       {
-        handle_keypress (ws, &xev->xkey);
+        handle_keypress (ws, &xev->xkey, filter_p);
         ws->caps_p = (xev->xkey.state & LockMask);
         if (ws->auth_state == AUTH_NOTIFY)
           ws->auth_state = AUTH_CANCEL;
@@ -2109,6 +2378,14 @@ handle_event (window_state *ws, XEvent *xev)
     refresh_p = True;
     break;
 
+  case KeyRelease:
+    if (ws->bs_timer)
+      {
+        XtRemoveTimeOut (ws->bs_timer);
+        ws->bs_timer = 0;
+      }
+    break;
+
   case ButtonPress:
   case ButtonRelease:
     {
@@ -2141,6 +2418,23 @@ cursor_timer (XtPointer closure, XtIntervalId *id)
 }
 
 
+/* Auto-repeat Backspace, since XInput2 doesn't do autorepeat. */
+static void
+bs_timer (XtPointer closure, XtIntervalId *id)
+{
+  window_state *ws = (window_state *) closure;
+  if (ws->bs_timer)
+    XtRemoveTimeOut (ws->bs_timer);
+  ws->bs_timer = 0;
+  /* Erase line */
+  memset (ws->plaintext_passwd, 0, sizeof (ws->plaintext_passwd));
+  memset (ws->plaintext_passwd_char_size, 0, 
+          sizeof (ws->plaintext_passwd_char_size));
+  memset (ws->censored_passwd, 0, sizeof(ws->censored_passwd));
+  window_draw (ws);
+}
+
+
 /* Redraw the window for the thermometer, and exit if the time has expired. 
  */
 static void
@@ -2156,6 +2450,19 @@ thermo_timer (XtPointer closure, XtIntervalId *id)
 }
 
 
+/* This is horrible, but if we don't ignore XInput events when running
+   test-passwd.c, then every mouse is delivered twice (both XI and Xlib)
+   so we only see every other page, because every click is a double-click.
+   Only test_auth_conv() calls this to set this flag.
+ */
+static Bool test_mode_ignore_xi_events = False;
+void
+xscreensaver_auth_test_mode (void)
+{
+  test_mode_ignore_xi_events = True;
+}
+
+
 static void
 gui_main_loop (window_state *ws, Bool splash_p, Bool notification_p)
 {
@@ -2199,6 +2506,7 @@ gui_main_loop (window_state *ws, Bool splash_p, Bool notification_p)
     {
       XEvent xev;
       XtInputMask m = XtAppPending (ws->app);
+      Bool filtered_p = False;
 
       if (m & XtIMXEvent)
         /* Process timers then block on an X event (which we know is there) */
@@ -2228,24 +2536,38 @@ gui_main_loop (window_state *ws, Bool splash_p, Bool notification_p)
       if ((m & ~XtIMXEvent) && !ws->splash_p)
         refresh_p = True;   /* In auth mode, all timers refresh */
 
-      if (verbose_p || debug_p)
-        print_xinput_event (ws->dpy, &xev, "");
-
       /* Convert XInput events to Xlib events, for simplicity and familiarity.
        */
-      if (xev.xcookie.type == GenericEvent &&
-          xev.xcookie.extension == ws->xi_opcode &&
-          (xev.xcookie.data || XGetEventData (ws->dpy, &xev.xcookie)))
-        {
-          XEvent ev2;
-          Bool ok =
-            xinput_event_to_xlib (xev.xcookie.evtype, xev.xcookie.data, &ev2);
-          XFreeEventData (ws->dpy, &xev.xcookie);
-          if (ok)
-            xev = ev2;
-        }
+      {
+        XEvent ev2;
+        Bool swap_p = False;
 
-      if (handle_event (ws, &xev))
+        if (xev.xcookie.type == GenericEvent &&
+            xev.xcookie.extension == ws->xi_opcode &&
+            (xev.xcookie.data || XGetEventData (ws->dpy, &xev.xcookie)))
+          {
+            swap_p = xinput_event_to_xlib (xev.xcookie.evtype,
+                                           xev.xcookie.data, &ev2);
+            if (test_mode_ignore_xi_events)
+              swap_p = False;
+          }
+
+        /* XFilterEvent does not work on XInput events, so check the
+           newly-synthesized X11 event instead. */
+        filtered_p = XFilterEvent ((swap_p ? &ev2 : &xev), ws->window);
+
+        /* Log the original event, not the synthetic one. */
+        if (verbose_p || debug_p)
+          print_xinput_event (ws->dpy, &xev, ws->ic,
+                              (filtered_p ? " (filtered)" : ""));
+        if (swap_p)
+          {
+            XFreeEventData (ws->dpy, &xev.xcookie);
+            xev = ev2;
+          }
+      }
+
+      if (handle_event (ws, &xev, filtered_p))
         refresh_p = True;
 
       XtDispatchEvent (&xev);
@@ -2277,6 +2599,7 @@ gui_main_loop (window_state *ws, Bool splash_p, Bool notification_p)
            keyboard layout would count as such.  It does not. */
         if (verbose_p)
           fprintf (stderr, "%s: MappingNotify\n", blurb());
+        XRefreshKeyboardMapping (&xev.xmapping);
         get_keyboard_layout (ws);
         refresh_p = True;
         break;
@@ -2296,6 +2619,7 @@ gui_main_loop (window_state *ws, Bool splash_p, Bool notification_p)
           XkbEvent *xkb = (XkbEvent *) &xev;
           if (verbose_p)
             fprintf (stderr, "%s: XKB event %d\n", blurb(), xkb->any.xkb_type);
+          /* Possibly the only event of interest here is XkbStateNotify = 2. */
           get_keyboard_layout (ws);
           refresh_p = True;
         }
@@ -2382,7 +2706,7 @@ window_state *global_ws = 0;
    to combine multiple messages onto a single dialog if PAM splits them
    between calls to this function.
 
-   Returns True on success.  If the user timed out or cancelled, we just exit.
+   Returns True on success.  If the user timed out or canceled, we just exit.
  */
 Bool
 xscreensaver_auth_conv (void *closure,
@@ -2550,10 +2874,10 @@ xscreensaver_auth_finished (void *closure, Bool authenticated_p)
 
 
 void
-xscreensaver_splash (void *closure)
+xscreensaver_splash (void *closure, Bool disable_settings_p)
 {
   Widget root_widget = (Widget) closure;
-  window_state *ws = window_init (root_widget, True);
+  window_state *ws = window_init (root_widget, disable_settings_p ? 2 : 1);
   ws->auth_state = AUTH_READ;
   gui_main_loop (ws, True, False);
   destroy_window (ws);
