@@ -1,4 +1,4 @@
-/* xscreensaver, Copyright © 1991-2021 Jamie Zawinski <jwz@jwz.org>
+/* xscreensaver, Copyright © 1991-2025 Jamie Zawinski <jwz@jwz.org>
  *
  * Permission to use, copy, modify, distribute, and sell this software and its
  * documentation for any purpose is hereby granted without fee, provided that
@@ -14,6 +14,6 @@
 
 extern Bool init_xinput (Display *dpy, int *opcode_ret);
 extern Bool xinput_event_to_xlib (int evtype, XIDeviceEvent *in, XEvent *out);
-extern void print_xinput_event (Display *, XEvent *, const char *desc);
+extern void print_xinput_event (Display *, XEvent *, XIC, const char *desc);
 
 #endif /* __XSCREENSAVER_XINPUT_H__ */
