@@ -1,4 +1,4 @@
-/* xscreensaver, Copyright © 1991-2021 Jamie Zawinski <jwz@jwz.org>
+/* xscreensaver, Copyright © 1991-2025 Jamie Zawinski <jwz@jwz.org>
  *
  * Permission to use, copy, modify, distribute, and sell this software and its
  * documentation for any purpose is hereby granted without fee, provided that
@@ -36,6 +36,10 @@
 #include <sys/types.h>
 #include <pwd.h>
 
+#ifndef HAVE_XINPUT
+# error The XInput2 extension is required
+#endif
+
 #include <X11/extensions/XInput2.h>
 
 #ifdef HAVE_RANDR
@@ -47,6 +51,11 @@
 # include <libintl.h>
 #endif
 
+#ifdef HAVE_WAYLAND
+# include "wayland-dpy.h"
+# include "wayland-dpms.h"
+#endif
+
 #include "xscreensaver.h"
 #include "version.h"
 #include "atoms.h"
@@ -86,16 +95,20 @@ maybe_reload_init_file (saver_info *si)
   saver_preferences *p = &si->prefs;
   if (init_file_changed_p (p))
     {
+      Bool ov = p->verbose_p;
       if (p->verbose_p)
 	fprintf (stderr, "%s: file \"%s\" has changed, reloading\n",
 		 blurb(), init_file_name());
 
       load_init_file (si->dpy, p);
 
+      if (ov)
+        p->verbose_p = True;
+
       /* If the DPMS settings in the init file have changed, change the
          settings on the server to match.  This also would have happened at
          the watchdog timer. */
-      sync_server_dpms_settings (si->dpy, p);
+      sync_server_dpms_settings (si);
     }
 }
 
@@ -133,13 +146,25 @@ saver_ehandler (Display *dpy, XErrorEvent *error)
 static void
 connect_to_server (saver_info *si)
 {
+  saver_preferences *p = &si->prefs;
   Widget toplevel_shell;
   Window daemon_window;
   XrmOptionDescRec options;
-  char *p;
+  char *name;
   int ac = 1;
   char *av[] = { "xscreensaver" };    /* For Xt and Xrm purposes */
 
+  /* Make sure the X11 socket doesn't get allocated to stderr: >&- 2>&-.
+     Parent "xscreensaver" already did this, but let's make sure. */
+  {
+    int fd0 = open ("/dev/null", O_RDWR);
+    int fd1 = open ("/dev/null", O_RDWR);
+    int fd2 = open ("/dev/null", O_RDWR);
+    if (fd0 > 2) close (fd0);
+    if (fd1 > 2) close (fd1);
+    if (fd2 > 2) close (fd2);
+  }
+
   XSetErrorHandler (saver_ehandler);
 
   toplevel_shell = XtAppInitialize (&si->app, progclass, &options, 0,
@@ -147,7 +172,7 @@ connect_to_server (saver_info *si)
 
   si->dpy = XtDisplay (toplevel_shell);
   si->prefs.db = XtDatabase (si->dpy);
-  XtGetApplicationNameAndClass (si->dpy, &p, &progclass);
+  XtGetApplicationNameAndClass (si->dpy, &name, &progclass);
 
   db = si->prefs.db;	/* resources.c needs this */
 
@@ -160,15 +185,63 @@ connect_to_server (saver_info *si)
   daemon_window = find_screensaver_window (si->dpy, 0);
   if (daemon_window)
     {
+      Window root = RootWindow (si->dpy, 0);  /* always screen 0 */
       XWindowAttributes xgwa;
       XGetWindowAttributes (si->dpy, daemon_window, &xgwa);
       XSelectInput (si->dpy, daemon_window,
                     xgwa.your_event_mask | PropertyChangeMask);
+
+      /* For tracking changes to XA_SCREENSAVER_STATUS on the root. */
+      XGetWindowAttributes (si->dpy, root, &xgwa);
+      XSelectInput (si->dpy, root,
+                    xgwa.your_event_mask | PropertyChangeMask);
     }
   else
     {
-      fprintf (stderr, "%s: xscreensaver does not seem to be running!\n",
-               blurb());
+      if (p->verbose_p)
+        fprintf (stderr, "%s: xscreensaver does not seem to be running!\n",
+                 blurb());
+
+      /* Under normal circumstances, that window should have been created by
+         the "xscreensaver" process.  But if for some reason someone has run
+         "xscreensaver-gfx" directly (Wayland swayidle?) we need this window
+         to exist for ClientMessages to be receivable.  So let's make one.
+       */
+      XClassHint class_hints;
+      XSetWindowAttributes attrs;
+      unsigned long attrmask = 0;
+      pid_t pid = getpid();
+      char *id;
+      const char *version_number = XSCREENSAVER_VERSION;
+
+      class_hints.res_name  = (char *) progname;  /* not const? */
+      class_hints.res_class = "XScreenSaver";
+      id = (char *) malloc (20);
+      sprintf (id, "%lu", (unsigned long) pid);
+    
+      attrmask = CWOverrideRedirect | CWEventMask;
+      attrs.override_redirect = True;
+      attrs.event_mask = PropertyChangeMask;
+
+      daemon_window = XCreateWindow (si->dpy, RootWindow (si->dpy, 0),
+                                     0, 0, 1, 1, 0,
+                                     DefaultDepth (si->dpy, 0), InputOutput,
+                                     DefaultVisual (si->dpy, 0),
+                                     attrmask, &attrs);
+      XStoreName (si->dpy, daemon_window, "XScreenSaver GFX");
+      XSetClassHint (si->dpy, daemon_window, &class_hints);
+      XChangeProperty (si->dpy, daemon_window, XA_WM_COMMAND, XA_STRING,
+                       8, PropModeReplace, (unsigned char *) progname,
+                       strlen (progname));
+      XChangeProperty (si->dpy, daemon_window, XA_SCREENSAVER_VERSION,
+                       XA_STRING, 8, PropModeReplace,
+                       (unsigned char *) version_number,
+                       strlen (version_number));
+      XChangeProperty (si->dpy, daemon_window, XA_SCREENSAVER_ID, XA_STRING,
+                       8, PropModeReplace, (unsigned char *) id, strlen (id));
+      free (id);
+
+      si->all_clientmessages_p = True;
     }
 }
 
@@ -220,27 +293,69 @@ read_status_prop (saver_info *si)
       ssi->current_hack = -1;
     }
 
-  XGetWindowProperty (si->dpy, w, XA_SCREENSAVER_STATUS,
-                      0, 999, False, XA_INTEGER, &type, &format, &nitems,
-                      &bytesafter, &dataP);
-  if (dataP && type == XA_INTEGER && nitems >= 3)
+  /* XA_SCREENSAVER_STATUS format documented in windows.c. */
+  if (XGetWindowProperty (si->dpy, w,
+                          XA_SCREENSAVER_STATUS,
+                          0, 999, False, XA_INTEGER,
+                          &type, &format, &nitems, &bytesafter,
+                          &dataP)
+      == Success
+      && type == XA_INTEGER
+      && nitems >= 3
+      && dataP)
     {
-      for (i = 2; i < nitems; i++)
+      PROP32 *data = (PROP32 *) dataP;
+      int off = 3;
+      for (i = off; i < nitems; i++)
         {
-          int j = i - 2;
+          int j = i - off;
           if (j < si->nscreens)
             {
               saver_screen_info *ssi = &si->screens[j];
-              int n = ((PROP32 *) dataP)[i];
-              ssi->current_hack =  n-1;  /* 1-based */
+              int n = data[i];
+              ssi->current_hack = n-1;  /* 1-based */
             }
         }
     }
+
   if (dataP)
     XFree (dataP);
 }
 
 
+/* When we get a PropertyChange event on XA_SCREENSAVER_STATUS, note
+   whether the auth dialog is currently posted.
+ */
+static void
+status_prop_changed (saver_info *si)
+{
+  Window w = RootWindow (si->dpy, 0);  /* always screen 0 */
+  Atom type;
+  unsigned char *dataP = 0;
+  int format;
+  unsigned long nitems, bytesafter;
+  Bool oauth_p = si->auth_p;
+
+  if (XGetWindowProperty (si->dpy, w,
+                          XA_SCREENSAVER_STATUS,
+                          0, 999, False, XA_INTEGER,
+                          &type, &format, &nitems, &bytesafter,
+                          &dataP)
+      == Success
+      && type == XA_INTEGER
+      && nitems >= 3
+      && dataP)
+    {
+      PROP32 *data = (PROP32 *) dataP;
+      si->auth_p = (data[0] == XA_AUTH);
+      if (si->auth_p != oauth_p)  /* Faster watchdog while authenticating */
+        reset_watchdog_timer (si);
+    }
+  else
+    si->auth_p = False;
+}
+
+
 /* Processing ClientMessage events.
    Both xscreensaver and xscreensaver-gfx handle these; some are handled
    exclusively by one program or another, and a couple (next, prev) are
@@ -304,6 +419,14 @@ handle_clientmessage (saver_info *si, XEvent *xev)
       si->selection_mode = which;
       goto CYCLE;
     }
+  else if (si->all_clientmessages_p)
+    {
+      /* The xscreensaver daemon is not running, so return an error response
+         for unhandled ClientMessages so that xscreensaver-command doesn't
+         have to time out waiting for a response that will not arrive. */
+      clientmessage_response (dpy, xev, False,
+                              "xscreensaver daemon not running");
+    }
   else
     {
       /* All other ClientMessages are handled by xscreensaver rather than
@@ -334,6 +457,16 @@ debug_multiscreen_timer (XtPointer closure, XtIntervalId *id)
 #endif /* DEBUG_MULTISCREEN */
 
 
+#ifdef HAVE_WAYLAND
+static void
+wayland_dpy_cb (XtPointer closure, int *source, XtInputId *id)
+{
+  saver_info *si = (saver_info *) closure;
+  wayland_dpy_process_events (si->wayland_dpy, False);
+}
+#endif /* HAVE_WAYLAND */
+
+
 static void
 main_loop (saver_info *si, Bool init_p)
 {
@@ -342,9 +475,33 @@ main_loop (saver_info *si, Bool init_p)
   if (init_p && p->verbose_p)
     print_available_extensions (si);
 
+# ifdef HAVE_WAYLAND
+  si->wayland_dpy = wayland_dpy_connect();
+  if (si->wayland_dpy)
+    {
+      /* Process input on the Wayland socket. */
+      int fd = wayland_dpy_get_fd (si->wayland_dpy);
+      XtAppAddInput (si->app, fd,
+                     (XtPointer) (XtInputReadMask | XtInputExceptMask),
+                     wayland_dpy_cb, (XtPointer) si);
+      si->wayland_dpms = wayland_dpms_init (si->wayland_dpy);
+    }
+# endif /* HAVE_WAYLAND */
+
   initialize_randr (si);
   update_screen_layout (si);
-  describe_monitor_layout (si->monitor_layout);
+
+  if (p->verbose_p)
+    describe_monitor_layout (si->monitor_layout);
+
+  /* The screen blanked (will blank) at xscreensaver-gfx's launch time.
+     The last user activity time is presumed to be 'timeout' before that.
+     Unless the command line included --next, --prev, --demo, etc.
+   */
+  si->blank_time = time ((time_t *) 0);
+  si->activity_time = (si->selection_mode == 0
+                       ? si->blank_time - p->timeout / 1000
+                       : si->blank_time);
 
   initialize_screensaver_window (si);
   init_sigchld (si);
@@ -381,6 +538,12 @@ main_loop (saver_info *si, Bool init_p)
 
       if (event.x_event.xany.type == ClientMessage)
         handle_clientmessage (si, &event.x_event);
+
+      else if (event.x_event.xany.type == PropertyNotify &&
+               event.x_event.xproperty.state == PropertyNewValue &&
+               event.x_event.xproperty.atom == XA_SCREENSAVER_STATUS)
+        status_prop_changed (si);
+
 # ifdef HAVE_RANDR
       else if (si->using_randr_extension &&
                (event.x_event.type == 
@@ -389,26 +552,26 @@ main_loop (saver_info *si, Bool init_p)
           /* The Resize and Rotate extension sends an event when the
              size, rotation, or refresh rate of any screen has changed.
            */
-          if (p->verbose_p)
-            {
-              int screen = XRRRootToScreen (si->dpy, event.xrr_event.window);
-              fprintf (stderr, "%s: %d: screen change event received\n",
-                       blurb(), screen);
-            }
+          Bool changed_p;
 
           /* Inform Xlib that it's ok to update its data structures. */
           XRRUpdateConfiguration (&event.x_event);
 
           /* Resize the existing xscreensaver windows and cached ssi data. */
-          if (update_screen_layout (si))
+          changed_p = update_screen_layout (si);
+
+          if (p->verbose_p)
             {
-              if (p->verbose_p)
-                {
-                  fprintf (stderr, "%s: new layout:\n", blurb());
-                  describe_monitor_layout (si->monitor_layout);
-                }
-              resize_screensaver_window (si);
+              int screen = XRRRootToScreen (si->dpy, event.xrr_event.window);
+              fprintf (stderr, "%s: %d: screen change event: %s\n",
+                       blurb(), screen,
+                       (changed_p ? "new layout:" : "layout unchanged"));
+              if (changed_p)
+                describe_monitor_layout (si->monitor_layout);
             }
+
+          if (changed_p)
+            resize_screensaver_window (si);
         }
 # endif /* HAVE_RANDR */
 
@@ -495,6 +658,13 @@ main (int argc, char **argv)
           dpy_str = argv[++i];
           if (!dpy_str) goto HELP;
         }
+      else if (!strcmp (argv[i], "-ver") ||
+               !strcmp (argv[i], "-vers") ||
+               !strcmp (argv[i], "-version"))
+        {
+          fprintf (stderr, "%s\n", screensaver_id+4);
+          exit (1);
+        }
       else if (!strcmp (argv[i], "-sync") ||
                !strcmp (argv[i], "-synch") ||
                !strcmp (argv[i], "-synchronize") ||
