@@ -1,4 +1,4 @@
-/* texfont, Copyright © 2005-2021 Jamie Zawinski <jwz@jwz.org>
+/* texfont, Copyright © 2005-2025 Jamie Zawinski <jwz@jwz.org>
  *
  * Permission to use, copy, modify, distribute, and sell this software and its
  * documentation for any purpose is hereby granted without fee, provided that
@@ -115,6 +115,8 @@ struct texture_font_data {
   XftFont *xftfont;
   int cache_size;
   texfont_cache *cache;
+  Bool dropshadow_p;
+  Bool mipmap_p;
 # ifdef HAVE_GLSL
   Bool shaders_initialized, use_shaders;
   GLuint shader_program;
@@ -137,7 +139,6 @@ bitmap_to_texture (const texture_font_data *tfdata, Pixmap p,
                    Visual *visual, int depth, int *wP, int *hP)
 {
   Display *dpy = tfdata->dpy;
-  Bool mipmap_p = True;
   int ow = *wP;
   int oh = *hP;
   GLsizei w2 = (GLsizei) to_pow2 (ow);
@@ -146,6 +147,8 @@ bitmap_to_texture (const texture_font_data *tfdata, Pixmap p,
   XImage *image = 0;
   unsigned char *data = (unsigned char *) calloc (w2 * 2, (h2 + 1));
   unsigned char *out = data;
+  GLint rowpack = 0;
+  GLint alignment = 0;
 
   /* OpenGLES doesn't support GL_INTENSITY, so instead of using a
      texture with 1 byte per pixel, the intensity value, we have
@@ -195,13 +198,8 @@ bitmap_to_texture (const texture_font_data *tfdata, Pixmap p,
     XGetSubImage (dpy, p, 0, 0, ow, oh, ~0L, ZPixmap, image, 0, 0);
   }
 
-# ifdef HAVE_JWZGLES
-  /* This would work, but it's wasteful for no benefit. */
-  mipmap_p = False;
-# endif
-
 # ifdef DUMP_BITMAPS
-  fprintf (stderr, "\n");
+  fprintf (stderr, "\n\n%d x %d => %d x %d, %d\n", ow, oh, w2, h2, scale);
 # endif
   for (y = 0; y < h2; y++) {
     for (x = 0; x < w2; x++) {
@@ -219,7 +217,7 @@ bitmap_to_texture (const texture_font_data *tfdata, Pixmap p,
       pixel = ((r >> 24) | (r >> 16) | (r >> 8) | r) & 0xFF;
 
 # ifdef DUMP_BITMAPS
-      if (sx < ow && sy < oh)
+      if (sx < ow && sy < oh && sx <= 79 && sy <= 40)
 #  ifdef HAVE_JWXYZ
         fprintf (stderr, "%c", 
                  r >= 0xFF000000 ? '#' : 
@@ -244,7 +242,7 @@ bitmap_to_texture (const texture_font_data *tfdata, Pixmap p,
       *out++ = pixel;
     }
 # ifdef DUMP_BITMAPS
-    fprintf (stderr, "\n");
+    if (y * scale <= 40) fprintf (stderr, "\n");
 # endif
   }
 
@@ -262,6 +260,12 @@ bitmap_to_texture (const texture_font_data *tfdata, Pixmap p,
 
   image = 0;
 
+  glGetIntegerv (GL_UNPACK_ROW_LENGTH, &rowpack);
+  glGetIntegerv (GL_UNPACK_ALIGNMENT, &alignment);
+
+  glPixelStorei (GL_UNPACK_ROW_LENGTH, 0);
+  glPixelStorei (GL_UNPACK_ALIGNMENT, 1);
+
   {
 # ifdef GL_INTENSITY
     GLuint iformat = GL_INTENSITY;
@@ -282,7 +286,7 @@ bitmap_to_texture (const texture_font_data *tfdata, Pixmap p,
     else
 #endif /* HAVE_GLSL */
       {
-        if (mipmap_p)
+        if (tfdata->mipmap_p)
           gluBuild2DMipmaps (GL_TEXTURE_2D, iformat, w2, h2, format, 
                              type, data);
         else
@@ -291,10 +295,13 @@ bitmap_to_texture (const texture_font_data *tfdata, Pixmap p,
       }
   }
 
+  glPixelStorei (GL_UNPACK_ROW_LENGTH, rowpack);
+  glPixelStorei (GL_UNPACK_ALIGNMENT, alignment);
+
   {
     char msg[100];
     sprintf (msg, "texture font %s (%d x %d)",
-             mipmap_p ? "gluBuild2DMipmaps" : "glTexImage2D",
+             tfdata->mipmap_p ? "gluBuild2DMipmaps" : "glTexImage2D",
              w2, h2);
     check_gl_error (msg);
   }
@@ -377,6 +384,21 @@ load_texture_font (Display *dpy, char *res)
   data->dpy = dpy;
   data->xftfont = f;
   data->cache_size = cache_size;
+  data->dropshadow_p =
+    !get_boolean_resource (dpy, "texFontOmitDropShadow", "Boolean");
+
+  data->mipmap_p = True;
+
+# if defined(__APPLE__) && !defined(HAVE_COCOA)   /* macOS X11 */
+  /* Some time before macOS 14.7.3, gluBuild2DMipmaps() started segfaulting. */
+  data->mipmap_p = False;
+# endif
+
+# ifdef HAVE_JWZGLES
+  /* This would work, but it's wasteful for no benefit. */
+  /* Wait, is it ever useful? */
+  data->mipmap_p = False;
+# endif
 
 #ifdef HAVE_GLSL
   /* Setting data->shaders_initialized to False will cause
@@ -764,7 +786,7 @@ enable_texture_string_parameters (texture_font_data *data)
 {
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER,
-                   GL_LINEAR_MIPMAP_LINEAR);
+                   data->mipmap_p ? GL_LINEAR_MIPMAP_LINEAR : GL_LINEAR);
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
   glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
   glEnable (GL_BLEND);
@@ -1088,6 +1110,7 @@ initialize_textfont_shaders_glsl (texture_font_data *data)
       data->tex_sampler_index != -1)
     {
       data->use_shaders = True;
+      data->mipmap_p = True;
       data->shaders_initialized = True;
     }
   else
@@ -1320,12 +1343,13 @@ print_texture_label (Display *dpy,
 
 # ifdef HAVE_GLSL
       if (data->use_shaders)
-        glUniform4f (data->font_color_index, 0, 0, 0, 1);
+        glUniform4f (data->font_color_index, 0, 0, 0, color[3]);
       else
 # endif /* HAVE_GLSL */
-        glColor3f (0, 0, 0);
+        glColor4f (0, 0, 0, color[3]);
 
-      for (i = 0; i < countof(offsets); i++)
+      for (i = (data->dropshadow_p ? 0 : countof(offsets)-1);
+           i < countof(offsets); i++)
         {
           if (offsets[i].x == 0)
             {
