@@ -1,4 +1,4 @@
-/* xscreensaver, Copyright (c) 1992-2018 Jamie Zawinski <jwz@jwz.org>
+/* xscreensaver, Copyright © 1992-2025 Jamie Zawinski <jwz@jwz.org>
  *
  * Permission to use, copy, modify, distribute, and sell this software and its
  * documentation for any purpose is hereby granted without fee, provided that
@@ -83,9 +83,11 @@
  */
 
 #include "utils.h"
-#include "grabscreen.h"
+#include "grabclient.h"
 #include "resources.h"
 #include "yarandom.h"
+#include "xft.h"
+#include "font-retry.h"
 
 #ifdef HAVE_JWXYZ
 # include "jwxyz.h"
@@ -97,6 +99,7 @@
 #endif /* !HAVE_COCOA */
 
 #include <sys/stat.h>
+#include <errno.h>
 
 #ifdef HAVE_UNISTD_H
 # include <unistd.h>
@@ -173,44 +176,6 @@ drawable_window_p (Display *dpy, Drawable d)
 }
 
 
-static Bool
-xscreensaver_window_p (Display *dpy, Window window)
-{
-  Atom type;
-  int format;
-  unsigned long nitems, bytesafter;
-  unsigned char *version;
-  if (XGetWindowProperty (dpy, window,
-			  XInternAtom (dpy, "_SCREENSAVER_VERSION", False),
-			  0, 1, False, XA_STRING,
-			  &type, &format, &nitems, &bytesafter,
-			  &version)
-      == Success
-      && type != None)
-    return True;
-  return False;
-}
-
-
-/* XCopyArea seems not to work right on SGI O2s if you draw in SubwindowMode
-   on a window whose depth is not the maximal depth of the screen?  Or
-   something.  Anyway, things don't work unless we: use SubwindowMode for
-   the real root window (or a legitimate virtual root window); but do not
-   use SubwindowMode for the xscreensaver window.  I make no attempt to
-   explain.
- */
-Bool
-use_subwindow_mode_p (Screen *screen, Window window)
-{
-  if (window != VirtualRootWindowOfScreen(screen))
-    return False;
-  else if (xscreensaver_window_p(DisplayOfScreen(screen), window))
-    return False;
-  else
-    return True;
-}
-
-
 static void
 checkerboard (Screen *screen, Drawable drawable)
 {
@@ -336,22 +301,22 @@ hack_subproc_environment (Display *dpy)
   /* Store $DISPLAY into the environment, so that the $DISPLAY variable that
      the spawned processes inherit is what we are actually using.
    */
+
+# if !(defined(__APPLE__) && !defined(HAVE_COCOA))  /* not macOS X11 */
+  /* XQuartz's weird $DISPLAY pathnames do not match DisplayString() */
+
   const char *odpy = DisplayString (dpy);
   char *ndpy = (char *) malloc(strlen(odpy) + 20);
   strcpy (ndpy, "DISPLAY=");
   strcat (ndpy, odpy);
-
-  /* Allegedly, BSD 4.3 didn't have putenv(), but nobody runs such systems
-     any more, right?  It's not Posix, but everyone seems to have it. */
-# ifdef HAVE_PUTENV
   if (putenv (ndpy))
     abort ();
-# endif /* HAVE_PUTENV */
 
   /* don't free (ndpy) -- some implementations of putenv (BSD 4.4,
      glibc 2.0) copy the argument, but some (libc4,5, glibc 2.1.2, MacOS)
      do not.  So we must leak it (and/or the previous setting). Yay.
    */
+# endif  /* not macOS X11 */
 }
 
 
@@ -505,7 +470,7 @@ xscreensaver_getimage_cb (XtPointer closure, int *fd, XtIntervalId *id)
 
 
 /* Loads an image into the Drawable.
-   When grabbing desktop images, the Window will be unmapped first.
+   When grabbing desktop images, the Window may be temporarily unmapped.
    Used only when running "real" X11, not jwxyz.
  */
 static void
@@ -546,9 +511,10 @@ load_random_image_x11 (Screen *screen, Window window, Drawable drawable,
 
   /* In case "cmd" fails, leave some random image on the screen, not just
      black or white, so that it's more obvious what went wrong. */
-  checkerboard (screen, drawable);
   if (window == drawable)
     print_loading_msg (screen, window);
+  else
+    checkerboard (screen, drawable);
 
   XSync (dpy, True);
   hack_subproc_environment (dpy);
@@ -620,8 +586,16 @@ xscreensaver_getimage_file_cb (XtPointer closure, int *source, XtInputId *id)
   char buf[10240];
   const char *dir = clo2->directory;
   char *absfile = 0;
+  int i = 0;
+
   *buf = 0;
-  fgets (buf, sizeof(buf)-1, clo2->pipe);
+  do {
+    errno = 0;
+    if (! fgets (buf, sizeof(buf)-1, clo2->pipe))
+      *buf = 0;
+  } while (errno == EINTR &&	/* fgets might fail due to SIGCHLD. */
+           i++ < 1000);		/* And just in case. */
+
   pclose (clo2->pipe);
   clo2->pipe = 0;
   XtRemoveInput (clo2->pipe_id);
@@ -829,6 +803,7 @@ load_random_image_cocoa (Screen *screen, Window window, Drawable drawable,
 
 # ifndef HAVE_IPHONE
   if (filep)
+    /* On macOS, get_string_resource() already expanded "~/". */
     dir = get_string_resource (dpy, "imageDirectory", "ImageDirectory");
 
   if (!dir || !*dir)
@@ -928,37 +903,36 @@ print_loading_msg (Screen *screen, Window window)
 {
   Display *dpy = DisplayOfScreen (screen);
   XWindowAttributes xgwa;
-  XGCValues gcv;
-  XFontStruct *f = 0;
-  GC gc;
-  char *fn = get_string_resource (dpy, "labelFont", "Font");
+  XftFont  *xftfont;
+  XftColor  xftcolor;
+  XftDraw  *xftdraw;
+  XGlyphInfo extents;
+  const char *fn = "sans-serif bold 36";
+  char *cname = get_string_resource (dpy, "foreground", "Foreground");
   const char *text = "Loading...";
   int w;
 
-  if (!fn) fn = get_string_resource (dpy, "titleFont", "Font");
-  if (!fn) fn = get_string_resource (dpy, "font", "Font");
-  if (!fn) fn = strdup ("-*-times-bold-r-normal-*-180-*");
-  f = XLoadQueryFont (dpy, fn);
-  if (!f) f = XLoadQueryFont (dpy, "fixed");
-  if (!f) abort();
-  free (fn);
-  fn = 0;
-
   XGetWindowAttributes (dpy, window, &xgwa);
-  w = XTextWidth (f, text, (int) strlen(text));
 
-  gcv.foreground = get_pixel_resource (dpy, xgwa.colormap,
-                                       "foreground", "Foreground");
-  gcv.background = get_pixel_resource (dpy, xgwa.colormap,
-                                       "background", "Background");
-  gcv.font = f->fid;
-  gc = XCreateGC (dpy, window, GCFont | GCForeground | GCBackground, &gcv);
-  XDrawImageString (dpy, window, gc,
-                    (xgwa.width - w) / 2,
-                    (xgwa.height - (f->ascent + f->descent)) / 2 + f->ascent,
-                    text, (int) strlen(text));
-  XFreeFont (dpy, f);
-  XFreeGC (dpy, gc);
+  xftfont = load_xft_font_retry (dpy, 0 /* screen_number(xgwa.screen) */, fn);
+
+  if (!cname) cname = strdup ("white");
+  XftColorAllocName (dpy, xgwa.visual, xgwa.colormap, cname, &xftcolor);
+  free (cname);
+
+  xftdraw = XftDrawCreate (dpy, window, xgwa.visual, xgwa.colormap);
+
+  XftTextExtentsUtf8 (dpy, xftfont, (FcChar8 *) text, strlen(text), &extents);
+  w = extents.xOff;
+
+  XftDrawStringUtf8 (xftdraw, &xftcolor, xftfont,
+                     (xgwa.width - w) / 2,
+                     (xgwa.height - (xftfont->ascent + xftfont->descent)) / 2 +
+                     xftfont->ascent,
+                     (FcChar8 *) text, strlen(text));
+
+  XftColorFree (dpy, xgwa.visual, xgwa.colormap, &xftcolor);
+  XftFontClose (dpy, xftfont);
   XSync (dpy, False);
 }
 
