@@ -1,4 +1,4 @@
-/* xscreensaver, Copyright © 1999-2021 Jamie Zawinski <jwz@jwz.org>
+/* xscreensaver, Copyright © 1999-2025 Jamie Zawinski <jwz@jwz.org>
  *
  * Permission to use, copy, modify, distribute, and sell this software and its
  * documentation for any purpose is hereby granted without fee, provided that
@@ -33,7 +33,8 @@
 # endif
 #endif /* HAVE_GL */
 
-#include "visual.h"		/* after EGL/egl.h */
+#include "visual.h"
+#include "visual-gl.h"		/* after EGL/egl.h */
 
 #undef countof
 #define countof(x) (sizeof(x)/sizeof(*(x)))
@@ -58,17 +59,20 @@ get_gl_visual (Screen *screen)
   int egl_major = -1, egl_minor = -1;
 
   /* This is re-used, no need to close it. */
-  egl_display = eglGetDisplay ((EGLNativeDisplayType) dpy);
+  egl_display = eglGetPlatformDisplay (EGL_PLATFORM_X11_KHR,
+                                       (void *) dpy, NULL);
   if (!egl_display)
     {
-      fprintf (stderr, "%s: eglGetDisplay failed\n", progname);
+      fprintf (stderr, "%s: eglGetPlatformDisplay failed\n", progname);
       return 0;
     }
 
   if (! eglInitialize (egl_display, &egl_major, &egl_minor))
     {
-      fprintf (stderr, "%s: eglInitialize failed\n", progname);
-      abort();
+      /* The library already printed this, but without progname:
+         "libEGL warning: DRI2: failed to create any config" */
+      /* fprintf (stderr, "%s: eglInitialize failed\n", progname); */
+      return 0;  /* We get here if running Xephyr in 8-bit pseudocolor */
     }
 
   /* Get the best EGL config on any visual, then see what visual it used. */
@@ -230,13 +234,13 @@ get_egl_config (Display *dpy, EGLDisplay *egl_display,
   };
   EGLint attrs[40];
   EGLint nconfig;
-  int i, j, k, iter, pass;
+  int i, i_start, j, k, iter, pass;
   Bool glslp;
 
-  i = 0;
+  i_start = 0;
 # ifdef SB
   if (! get_boolean_resource (dpy, "multiSample", "MultiSample"))
-    i = SB_COUNT;  /* skip over the multibuffer entries in 'attrs' */
+    i_start = SB_COUNT;  /* skip over the multibuffer entries in 'attrs' */
 # endif /* SB */
 
   glslp = get_boolean_resource (dpy, "prefersGLSL", "PrefersGLSL");
@@ -245,7 +249,7 @@ get_egl_config (Display *dpy, EGLDisplay *egl_display,
   *egl_config = 0;
   for (pass = 0; pass < iter; pass++)
     {
-      for (; i < countof(templates); i++)
+      for (i = i_start ; i < countof(templates); i++)
         {
           for (j = 0, k = 0; templates[i][j] != EGL_NONE; j += 2)
             {
@@ -284,8 +288,9 @@ get_egl_config (Display *dpy, EGLDisplay *egl_display,
 
   if (! *egl_config)
     {
-      fprintf (stderr, "%s: eglChooseConfig: no configs found\n", progname);
-      abort();
+      fprintf (stderr, "%s: no matching EGL config for X11 visual 0x%lx\n",
+               progname, (unsigned long) x11_visual_id);
+      return;
     }
 }
 #endif /* HAVE_EGL */
@@ -321,7 +326,7 @@ describe_gl_visual (FILE *f, Screen *screen, Visual *visual,
 
     /* This is re-used, no need to close it. */
     egl_display = eglGetPlatformDisplay (EGL_PLATFORM_X11_KHR,
-                                         (EGLNativeDisplayType) dpy, NULL);
+                                         (void *) dpy, NULL);
     if (!egl_display)
       {
         fprintf (stderr, "%s: eglGetPlatformDisplay failed\n", progname);
@@ -384,12 +389,15 @@ describe_gl_visual (FILE *f, Screen *screen, Visual *visual,
           else if (fields[i].i == EGL_TRANSPARENT_RED_VALUE && tt != EGL_NONE)
             sprintf (s, "%d, %d, %d", tr, tg, tb);
           else if (fields[i].i == EGL_CONFIG_CAVEAT)
-            strcpy (s, (v == EGL_NONE ? "none" :
-                        v == EGL_SLOW_CONFIG ? "slow" :
+            {
+              const char *s2 = (v == EGL_NONE ? "none" :
+                                v == EGL_SLOW_CONFIG ? "slow" :
 # ifdef EGL_NON_CONFORMANT
-                        v == EGL_NON_CONFORMANT ? "non-conformant" :
+                                v == EGL_NON_CONFORMANT ? "non-conformant" :
 # endif
-                        "???"));
+                                "???");
+              strcpy (s, s2);
+            }
           else if (fields[i].i == EGL_COLOR_BUFFER_TYPE)
             strcpy (s, (v == EGL_RGB_BUFFER ? "RGB" :
                         v == EGL_LUMINANCE_BUFFER ? "luminance" :
@@ -478,7 +486,7 @@ describe_gl_visual (FILE *f, Screen *screen, Visual *visual,
                (value == GLX_NONE_EXT ? "none" :
                 value == GLX_SLOW_VISUAL_EXT ? "slow" :
                 value == GLX_NON_CONFORMANT_EXT ? "non-conformant" :
-                "???");
+                "???"));
 #   else      
       fprintf (f, "    GLX rating:        %s\n",
                (value == GLX_NONE_EXT ? "none" :
@@ -597,3 +605,251 @@ validate_gl_visual (FILE *out, Screen *screen, const char *window_desc,
 
 #endif  /* !HAVE_GL */
 }
+
+
+# ifndef HAVE_EGL
+/* Gag -- we use this to turn X errors from glXCreateContext() into
+   something that will actually make sense to the user.
+ */
+static XErrorHandler orig_ehandler = 0;
+static Bool got_error = 0;
+
+static int
+BadValue_ehandler (Display *dpy, XErrorEvent *error)
+{
+  if (error->error_code == BadValue)
+    {
+      got_error = True;
+      return 0;
+    }
+  else
+    return orig_ehandler (dpy, error);
+}
+#endif /* !HAVE_EGL */
+
+
+/* Initializes OpenGL and returns a GLXContext or egl_data for the window.
+   The window is assumed to have been created with the proper visual.
+ */
+#ifdef HAVE_EGL
+egl_data  *openGL_context_for_window (Screen *screen, Window window)
+#else
+GLXContext openGL_context_for_window (Screen *screen, Window window)
+#endif
+{
+  Display *dpy = DisplayOfScreen (screen);
+  XWindowAttributes xgwa;
+  Visual *visual;
+  XVisualInfo vi_in, *vi_out;
+  int out_count;
+
+# ifdef HAVE_EGL
+  egl_data *egl_data_ret = NULL;
+# else
+  GLXContext glx_context = NULL;
+# endif
+
+  XGetWindowAttributes (dpy, window, &xgwa);
+  visual = xgwa.visual;
+
+  vi_in.screen = screen_number (screen);
+  vi_in.visualid = XVisualIDFromVisual (visual);
+  vi_out = XGetVisualInfo (dpy, VisualScreenMask|VisualIDMask,
+			   &vi_in, &out_count);
+  if (! vi_out) abort ();
+
+# ifdef HAVE_EGL
+  {
+    egl_data *d = (egl_data *) calloc (1, sizeof(*d));
+
+    /* The correct EGL config has been selected by calling get_egl_config()
+       from get_gl_visual and returning the corresponding X11 Visual.
+       That visual is the one that was used to create the window. We will
+       pass the corresponding visual ID to get_egl_config() to obtain the
+       same configuration here. */
+    unsigned int vid = XVisualIDFromVisual (visual);
+
+    const EGLint ctxattr1[] = {
+# ifdef HAVE_JWZGLES
+      EGL_CONTEXT_MAJOR_VERSION, 1,	/* Request an OpenGL ES 1.1 context. */
+      EGL_CONTEXT_MINOR_VERSION, 1,
+# else
+      EGL_CONTEXT_MAJOR_VERSION, 1,	/* Request an OpenGL 1.3 context. */
+      EGL_CONTEXT_MINOR_VERSION, 3,
+# endif
+      EGL_NONE
+    };
+    const EGLint *ctxattr = ctxattr1;
+
+# ifdef HAVE_GLES3
+    const EGLint ctxattr3[] = {
+      EGL_CONTEXT_MAJOR_VERSION, 3,	/* Request an OpenGL ES 3.0 context. */
+      EGL_CONTEXT_MINOR_VERSION, 0,
+      EGL_NONE
+    };
+
+    if (get_boolean_resource (dpy, "prefersGLSL", "PrefersGLSL"))
+      ctxattr = ctxattr3;
+# endif /* HAVE_GLES3 */
+
+    /* This is re-used, no need to close it. */
+    d->egl_display = eglGetPlatformDisplay (EGL_PLATFORM_X11_KHR,
+                                            (void *) dpy, NULL);
+    if (!d->egl_display)
+      {
+        fprintf (stderr, "%s: eglGetPlatformDisplay failed\n", progname);
+        abort();
+      }
+
+    get_egl_config (dpy, d->egl_display, vid, &d->egl_config);
+    if (!d->egl_config)
+      {
+        /* get_egl_config already printed this:
+        fprintf (stderr, "%s: no matching EGL config for X11 visual 0x%lx\n",
+                 progname, vi_out->visualid); */
+        /* returning 0 might be reasonable, but makes all the GL hacks
+           simply draw nothing in a loop. */
+        exit (1);
+      }
+
+    d->egl_surface = eglCreatePlatformWindowSurface (d->egl_display,
+                                                     d->egl_config,
+                                                     &window, NULL);
+    if (! d->egl_surface)
+      {
+        fprintf (stderr, "%s: eglCreatePlatformWindowSurface failed:"
+                 " window 0x%lx visual 0x%x\n", progname, window, vid);
+        abort();
+      }
+
+#ifdef HAVE_JWZGLES
+    /* This call is not strictly necessary to get an OpenGL ES context
+       since the default API is EGL_OPENGL_ES_API, but it  makes our
+       intention clear.
+     */
+    if (!eglBindAPI (EGL_OPENGL_ES_API))
+      {
+        fprintf (stderr, "%s: eglBindAPI failed\n", progname);
+      }
+#else /* !HAVE_JWZGLES */
+    /* This is necessary to get a OpenGL context instead of an OpenGLES
+       context.
+     */
+    if (!eglBindAPI (EGL_OPENGL_API))
+      {
+        fprintf (stderr, "%s: eglBindAPI failed\n", progname);
+      }
+#endif /* !HAVE_JWZGLES */
+
+    d->egl_context = eglCreateContext (d->egl_display, d->egl_config,
+                                       EGL_NO_CONTEXT, ctxattr);
+
+# ifdef HAVE_GLES3
+    /* If creation of a GLES 3.0 context failed, fall back to GLES 1.x. */
+    if (!d->egl_context && ctxattr != ctxattr1)
+      {
+        /* fprintf (stderr, "%s: eglCreateContext 3.0 failed\n", progname); */
+        d->egl_context = eglCreateContext (d->egl_display, d->egl_config,
+                                           EGL_NO_CONTEXT, ctxattr1);
+      }
+# endif /* HAVE_GLES3 */
+
+    if (!d->egl_context)
+      {
+        fprintf (stderr, "%s: eglCreateContext failed\n", progname);
+        abort();
+      }
+
+    /* describe_gl_visual (stderr, screen, visual, False); */
+
+    if (! eglMakeCurrent (d->egl_display, d->egl_surface, d->egl_surface,
+                          d->egl_context))
+      abort();
+
+    egl_data_ret = d;
+  }
+# else /* GLX */
+  {
+    XSync (dpy, False);
+    orig_ehandler = XSetErrorHandler (BadValue_ehandler);
+    glx_context = glXCreateContext (dpy, vi_out, 0, GL_TRUE);
+    XSync (dpy, False);
+    XSetErrorHandler (orig_ehandler);
+    if (got_error)
+      glx_context = 0;
+  }
+
+  if (!glx_context)
+    {
+      fprintf(stderr, "%s: couldn't create GL context for visual 0x%x.\n",
+	      progname, (unsigned int) XVisualIDFromVisual (visual));
+      exit(1);
+    }
+
+  glXMakeCurrent (dpy, window, glx_context);
+
+  {
+    GLboolean rgba_mode = 0;
+    glGetBooleanv(GL_RGBA_MODE, &rgba_mode);
+    if (!rgba_mode)
+      {
+	glIndexi (WhitePixelOfScreen (screen));
+	glClearIndex (BlackPixelOfScreen (screen));
+      }
+  }
+# endif /* GLX */
+
+  XFree((char *) vi_out);
+
+
+  /* jwz: the doc for glDrawBuffer says "The initial value is GL_FRONT
+     for single-buffered contexts, and GL_BACK for double-buffered
+     contexts."  However, I find that this is not always the case,
+     at least with Mesa 3.4.2 -- sometimes the default seems to be
+     GL_FRONT even when glGet(GL_DOUBLEBUFFER) is true.  So, let's
+     make sure.
+
+     Oh, hmm -- maybe this only happens when we are re-using the
+     xscreensaver window, and the previous GL hack happened to die with
+     the other buffer selected?  I'm not sure.  Anyway, this fixes it.
+   */
+  {
+    GLboolean d = False;
+    glGetBooleanv (GL_DOUBLEBUFFER, &d);
+    if (d)
+      glDrawBuffer (GL_BACK);
+    else
+      glDrawBuffer (GL_FRONT);
+  }
+
+  /* Sometimes glDrawBuffer() throws "invalid op". Dunno why. Ignore. */
+  while (glGetError() != GL_NO_ERROR)
+    ;
+
+# ifdef HAVE_EGL
+  return egl_data_ret;
+# else
+  return glx_context;
+# endif
+}
+
+#ifdef HAVE_EGL
+
+void
+openGL_destroy_context (Display *dpy, egl_data *data)
+{
+  eglDestroyContext (data->egl_display, data->egl_context);
+  eglDestroySurface (data->egl_display, data->egl_surface);
+  eglTerminate (data->egl_display);
+  free (data);
+}
+
+#else /* HAVE_EGL */
+
+void
+openGL_destroy_context (Display *dpy, GLXContext ctx)
+{
+  glXDestroyContext (dpy, ctx);
+}
+
+#endif /* HAVE_EGL */
